# -*- coding: utf-8 -*-
import os
import subprocess
import sys

NARPKG_PATH = os.path.dirname(__file__)


def _program(name: str, args: list[str]) -> int:
    return subprocess.call([os.path.join(NARPKG_PATH, name)] + args, close_fds=False)


def NAR():
    raise SystemExit(_program("NARexe", sys.argv[1:]))



from unittest.mock import Mock, call, patch

from narpyn.ona.nar import (
    expect_output,
    get_output,
    get_raw_output,
    send_input,
    setup_nars,
    setup_nars_ops,
)


@patch("subprocess.Popen")
def test_send_input(mock_popen):
    mock_stdin = Mock()
    mock_stdout = Mock()
    mock_popen.return_value = Mock(stdin=mock_stdin, stdout=mock_stdout)
    send_input(mock_popen(), "test_input")
    mock_stdin.write.assert_called_once_with("test_input\n")
    mock_stdin.flush.assert_called_once()


@patch("subprocess.Popen")
def test_get_raw_output(mock_popen):
    mock_stdin = Mock()
    mock_stdout = Mock()
    mock_popen.return_value = Mock(stdin=mock_stdin, stdout=mock_stdout)
    mock_stdout.readline.side_effect = [
        "^op1 args [a b c]\n",
        "done with 0 additional inference steps.\n",
    ]
    output = get_raw_output(mock_popen())
    assert len(output) == 1
    assert output[0] == "^op1 args [a b c]"


@patch("subprocess.Popen")
def test_get_output(mock_popen):
    mock_stdin = Mock()
    mock_stdout = Mock()
    mock_popen.return_value = Mock(stdin=mock_stdin, stdout=mock_stdout)
    mock_stdout.readline.side_effect = [
        "Input: task1! :|: occurrenceTime=5\n",
        "Derived: task2! :|: occurrenceTime=5\n",
        "Revised: sent. :|: occurrenceTime=2\n",
        "Answer: ^op1 ({SELF} * arg)! :|: occurrenceTime=5\n",
        "^op1 executed with args ({SELF} * arg)\n",
        "done with 0 additional inference steps.\n",
    ]
    output = get_output(mock_popen())
    assert len(output["input"]) == 1
    assert output["input"][0] == {
        "occurrenceTime": "5",
        "punctuation": "!",
        "term": "task1",
    }
    assert len(output["derivations"]) == 2
    assert output["derivations"][0] == {
        "occurrenceTime": "5",
        "punctuation": "!",
        "term": "task2",
    }
    assert output["derivations"][1] == {
        "occurrenceTime": "2",
        "punctuation": ".",
        "term": "sent",
    }
    assert len(output["answers"]) == 1
    assert output["answers"][0] == {
        "occurrenceTime": "5",
        "punctuation": "!",
        "term": "^op1 ({SELF} * arg)",
    }
    assert len(output["executions"]) == 1
    assert output["executions"][0] == {
        "operator": "^op1",
        "arguments": ["{SELF}", "arg"],
    }
    assert (
        output["raw"]
        == "Input: task1! :|: occurrenceTime=5\nDerived: task2! :|: occurrenceTime=5\nRevised: sent. :|: occurrenceTime=2\nAnswer: ^op1 ({SELF} * arg)! :|: occurrenceTime=5\n^op1 executed with args ({SELF} * arg)"
    )


@patch("subprocess.Popen")
def test_expect_output(mock_popen):
    mock_stdin = Mock()
    mock_stdout = Mock()
    mock_popen.return_value = Mock(stdin=mock_stdin, stdout=mock_stdout)
    mock_stdout.readline.side_effect = [
        "Input: task1. :|: occurrenceTime=now\n",
        "Derived: task2. occurrenceTime=eternal\n",
        "Revised: task3. :|: occurrenceTime=now\n",
        "Answer: task4. :|: occurrenceTime=now\n",
        "^op1 executed with args arg\n",
        "done with 0 additional inference steps.\n",
    ]
    targets = ["^op1"]
    output = expect_output(mock_popen(), targets)
    assert len(output["input"]) == 1
    assert output["input"][0] == {
        "occurrenceTime": "now",
        "punctuation": ".",
        "term": "task1",
    }
    assert len(output["derivations"]) == 2
    assert output["derivations"][0] == {
        "occurrenceTime": "eternal",
        "punctuation": ".",
        "term": "task2",
    }
    assert output["derivations"][1] == {
        "occurrenceTime": "now",
        "punctuation": ".",
        "term": "task3",
    }
    assert len(output["answers"]) == 1
    assert output["answers"][0] == {
        "occurrenceTime": "now",
        "punctuation": ".",
        "term": "task4",
    }
    assert len(output["executions"]) == 1
    assert output["executions"][0] == {"operator": "^op1", "arguments": ["arg"]}
    assert (
        output["raw"]
        == "Input: task1. :|: occurrenceTime=now\nDerived: task2. occurrenceTime=eternal\nRevised: task3. :|: occurrenceTime=now\nAnswer: task4. :|: occurrenceTime=now\n^op1 executed with args arg"
    )


@patch("subprocess.Popen")
def test_setup_nars_ops(mock_popen):
    mock_stdin = Mock()
    mock_stdout = Mock()
    mock_popen.return_value = Mock(stdin=mock_stdin, stdout=mock_stdout)
    ops = ["op1", "op2"]
    setup_nars_ops(mock_popen(), ops)
    mock_stdin.write.assert_has_calls(
        [
            call("*setopname 1 op1\n"),
            call("*setopname 2 op2\n"),
            call(f"*babblingops={len(ops)}\n"),
        ]
    )
    mock_stdin.flush.assert_has_calls([call()] * 3)
    send_input_calls = [call(mock_popen(), f"*babblingops={len(ops)}")]
    setup_nars_ops(mock_popen(), ops, babblingops=5)
    send_input_calls.append(call(mock_popen(), "*babblingops=5"))
    mock_stdin.write.assert_has_calls(
        [
            call("*setopname 1 op1\n"),
            call("*setopname 2 op2\n"),
            call("*babblingops=5\n"),
        ]
    )
    mock_stdin.flush.assert_has_calls([call()] * 3)
    assert mock_stdin.write.call_count == 6


@patch("subprocess.Popen")
def test_setup_nars(mock_popen):
    mock_stdin = Mock()
    mock_stdout = Mock()
    mock_popen.return_value = Mock(stdin=mock_stdin, stdout=mock_stdout)
    setup_nars(
        mock_popen(),
        ["op1", "op2"],
        motor_babbling=0.5,
        babblingops=3,
        volume=100,
        decision_threshold=0.8,
    )
    mock_stdin.write.assert_has_calls(
        [
            call("*reset\n"),
            call("*setopname 1 op1\n"),
            call("*setopname 2 op2\n"),
            call("*babblingops=3\n"),
            call("*motorbabbling=0.5\n"),
            call("*volume=100\n"),
            call("*decisionthreshold=0.8\n"),
        ]
    )
    mock_stdin.flush.assert_has_calls([call()] * 7)
    assert mock_stdin.write.call_count == 7



from narpyn.ona.narsese import (
    Goal,
    ext,
    loc,
    nal_demand,
    nal_now,
    narsify,
    parse_execution,
    parse_reason,
    parse_task,
    parse_truth_value,
    pos,
)


# Test narsify function
def test_narsify():
    assert narsify([1, 2, 3]) == "1,2,3"
    assert narsify(["a", "b", "c"]) == "a,b,c"
    assert narsify([]) == ""


# Test loc function
def test_loc():
    assert loc((0, 0)) == "loc_x0_y0"
    assert loc((1, 2)) == "loc_x1_y2"
    assert loc((-1, -2)) == "loc_x-1_y-2"


# Test pos function
def test_pos():
    assert pos("loc_x0_y0") == (0, 0)
    assert pos("loc_x1_y2") == (1, 2)
    assert pos("loc_x-1_y-2") == (-1, -2)


# Test ext function
def test_ext():
    assert ext("abc") == "{abc}"
    assert ext("def") == "{def}"
    assert ext("") == "{}"


# Test nal_demand function
def test_nal_demand():
    assert nal_demand("demand") == "demand! :|:"
    assert nal_demand("request") == "request! :|:"
    assert nal_demand("order") == "order! :|:"


# Test nal_now function
def test_nal_now():
    assert nal_now("statement") == "statement. :|:"
    assert nal_now("fact") == "fact. :|:"
    assert nal_now("truth") == "truth. :|:"


# Test parse_truth_value function
def test_parse_truth_value():
    assert parse_truth_value("frequency=0.5 confidence=0.9") == {
        "frequency": 0.5,
        "confidence": 0.9,
    }
    assert parse_truth_value("frequency=0.3, confidence=0.8") == {
        "frequency": 0.3,
        "confidence": 0.8,
    }


# Test parse_task function
def test_parse_task():
    assert parse_task("task! :|: occurrenceTime=3 Truth: f=0.7 c=0.89") == {
        "occurrenceTime": "3",
        "punctuation": "!",
        "term": "task",
        "truth": {"frequency": 0.7, "confidence": 0.89},
    }
    assert parse_task(
        "task! occurrenceTime=eternal Priority=1.0 Truth: f=0.5 c=0.9"
    ) == {
        "occurrenceTime": "eternal",
        "punctuation": "!",
        "term": "task",
        "truth": {"frequency": 0.5, "confidence": 0.9},
    }


# Test parse_reason function
def test_parse_reason():
    assert parse_reason(
        "implication: asdf. Truth: frequency=0.5 confidence=0.9 precondition: bls. :|: occurrenceTime=7 decision expectation=0.67"
    ) == {
        "desire": "0.67",
        "hypothesis": {
            "occurrenceTime": "eternal",
            "punctuation": ".",
            "term": "asdf",
            "truth": {"frequency": 0.5, "confidence": 0.9},
        },
        "precondition": {"occurrenceTime": "7", "punctuation": ".", "term": "bls"},
    }


# Test parse_execution function
def test_parse_execution():
    assert parse_execution("^op1 executed") == {"operator": "^op1", "arguments": []}
    assert parse_execution("^op2 executed with args ({SELF} * arg * x)") == {
        "operator": "^op2",
        "arguments": ["{SELF}", "arg", "x"],
    }


# Test Goal class
def test_goal():
    def is_satisfied():
        return True

    goal = Goal("symbol", is_satisfied, ["knowledge"])
    assert goal.symbol == "symbol"
    assert goal.satisfied() is True
    assert goal.knowledge == ["knowledge"]
    assert str(goal) == "Goal symbol; ['knowledge']"



from dataclasses import dataclass
from typing import Any, Callable, Optional


def narsify(observation):
    """Convert a Gym observation to NARS input"""
    return ",".join(str(x) for x in observation)


def loc(pos: tuple[int, int]) -> str:
    """Turn coordinates into a location string"""
    return f"loc_x{pos[0]}_y{pos[1]}"


def pos(loc: str) -> tuple[int, int]:
    """Turn a location string into coordinates"""
    coord_str = loc[4:]
    x, y = coord_str.split("_")
    x, y = x[1:], y[1:]
    return int(x), int(y)


def ext(s: str) -> str:
    """Just a helper to wrap strings in '{}'"""
    return "{" + s + "}"


def nal_demand(s: str) -> str:
    """Return NAL for demanding something"""
    return f"{s}! :|:"


def nal_now(s: str) -> str:
    """Return NAL statement in the present"""
    return f"{s}. :|:"


def parse_truth_value(tv_str: str) -> dict[str, float]:
    splits = tv_str.split(" ")
    freq_s, conf_s = splits[0], splits[1]
    if freq_s[-1] == ",":
        freq_s = freq_s[:-1]

    frequency = float(freq_s.split("=")[1])
    confidence = float(conf_s.split("=")[1])
    return {
        "frequency": frequency,
        "confidence": confidence,
    }


def parse_task(s: str) -> dict[str, Any]:
    m: dict[str, Any] = {"occurrenceTime": "eternal"}
    if " :|:" in s:
        m["occurrenceTime"] = "now"
        s = s.replace(" :|:", "")
        if "occurrenceTime" in s:
            m["occurrenceTime"] = s.split("occurrenceTime=")[1].split(" ")[0]
    sentence = (
        s.split(" occurrenceTime=")[0]
        if " occurrenceTime=" in s
        else s.split(" Priority=")[0]
    )
    m["punctuation"] = sentence[-4] if ":|:" in sentence else sentence[-1]
    m["term"] = (
        sentence.split(" creationTime")[0]
        .split(" occurrenceTime")[0]
        .split(" Truth")[0][:-1]
    )
    if "Truth" in s:
        m["truth"] = parse_truth_value(s.split("Truth: ")[1])
    return m


def parse_reason(sraw: str) -> Optional[dict[str, str]]:
    if "implication: " not in sraw:
        return None
    implication = parse_task(
        sraw.split("implication: ")[-1].split("precondition: ")[0]
    )  # last reason only (others couldn't be associated currently)
    precondition = parse_task(sraw.split("precondition: ")[-1].split("\n")[0])
    implication["occurrenceTime"] = "eternal"
    precondition["punctuation"] = implication["punctuation"] = "."
    reason: dict[str, Any] = {}
    reason["desire"] = sraw.split("decision expectation=")[-1].split(" ")[0]
    reason["hypothesis"] = implication
    reason["precondition"] = precondition
    return reason


def parse_execution(e: str) -> dict[str, Any]:
    op = e.split(" ")[0]
    argstr = e.split("args ")[1] if "args " in e else None
    if argstr is None:
        return {"operator": op, "arguments": []}
    return {
        "operator": op,
        "arguments": argstr[1:-1].split(" * ") if "*" in argstr else [argstr],
    }


@dataclass
class Goal:
    """A goal"""

    symbol: str
    satisfied: Callable
    knowledge: Optional[list[str]] = None

    def __repr__(self) -> str:
        return f"Goal {self.symbol}; {self.knowledge}"



import logging
import subprocess

from narpyn.ona.narsese import *

logger = logging.getLogger(__name__)


def send_input(process: subprocess.Popen, input_: str) -> None:
    """Send input to NARS server"""
    stdin, stdout = process.stdin, process.stdout
    if stdin is None or stdout is None:
        raise RuntimeError("Process has no stdin or stdout.")

    stdin.write(input_ + "\n")
    stdin.flush()


def get_raw_output(process: subprocess.Popen) -> list[str]:
    """Get raw output from NARS server"""
    stdin, stdout = process.stdin, process.stdout
    if stdin is None or stdout is None:
        raise RuntimeError("Process has no stdin or stdout.")

    stdin.write("0\n")
    stdin.flush()
    ret: str = ""
    before: list[str] = []
    while "done with 0 additional inference steps." != ret.strip():
        if ret != "":
            before.append(ret.strip())
        ret = stdout.readline()
    logger.debug("\n".join(before))
    return before


def get_output(process: subprocess.Popen) -> dict[str, Any]:
    lines = get_raw_output(process)
    executions = [parse_execution(l) for l in lines if l.startswith("^")]
    inputs = [
        parse_task(l.split("Input: ")[1]) for l in lines if l.startswith("Input:")
    ]
    derivations = [
        parse_task(l.split("Derived: " if l.startswith("Derived:") else "Revised: ")[1])
        for l in lines
        if l.startswith("Derived:") or l.startswith("Revised:")
    ]
    answers = [
        parse_task(l.split("Answer: ")[1]) for l in lines if l.startswith("Answer:")
    ]
    reason = parse_reason("\n".join(lines))
    return {
        "input": inputs,
        "derivations": derivations,
        "answers": answers,
        "executions": executions,
        "reason": reason,
        "raw": "\n".join(lines),
    }


def expect_output(
    process: subprocess.Popen,
    targets: list[str],
    think_ticks: int = 5,
    patience: int = 3,
    goal_reentry: Optional[Goal] = None,
) -> Optional[dict[str, Any]]:
    # TODO: refactor - this is basically for dealing with executions
    output = get_output(process)
    while not any(
        target in exe["operator"] for target in targets for exe in output["executions"]
    ):
        if patience <= 0:
            print("Patience has run out, returning None.")
            return None
        patience -= 1

        if goal_reentry is not None:
            send_input(process, nal_demand(goal_reentry.symbol))

        send_input(process, str(think_ticks))
        output = get_output(process)
    # ic("Got a matching output.")
    return output


def setup_nars_ops(
    process: subprocess.Popen, ops: list[str], babblingops: Optional[int] = None
) -> None:
    """Setup NARS operations"""
    for i, op in enumerate(ops):
        send_input(process, f"*setopname {i+1} {op}")
    if babblingops is None:
        send_input(process, f"*babblingops={len(ops)}")
    else:
        send_input(process, f"*babblingops={babblingops}")


def setup_nars(
    process: subprocess.Popen,
    ops: list[str],
    motor_babbling: Optional[float] = None,
    babblingops: Optional[int] = None,
    volume: Optional[int] = None,
    decision_threshold: Optional[float] = None,
) -> None:
    """Send NARS settings"""
    send_input(process, "*reset")
    setup_nars_ops(process, ops, babblingops=babblingops)
    if motor_babbling is not None:
        send_input(process, f"*motorbabbling={motor_babbling}")
    if volume is not None:
        send_input(process, f"*volume={volume}")
    if decision_threshold is not None:
        send_input(process, f"*decisionthreshold={decision_threshold}")


def start_nars(
    ops: list[str],
    motor_babbling: Optional[float] = None,
    babblingops: Optional[int] = None,
    volume: Optional[int] = None,
    decision_threshold: Optional[float] = None,
) -> subprocess.Popen:
    """Start an ONA process and set it up

    Returns: process: subprocess.Popen: the object of the spawned subprocess.
    """
    # Start the ONA process
    process_cmd = [
        "NAR",
        "shell",
    ]
    process: subprocess.Popen = subprocess.Popen(
        process_cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )

    # Configure ONA and return the process
    setup_nars(process, ops, motor_babbling, babblingops, volume, decision_threshold)
    return process
